/* 
 * misccmd.c --
 *
 *	This file contains code that implements miscellaneous
 *	xop subcommands, i.e. those that don't fall neatly
 *	into some other category.
 */

/*
 * Copyright (C) 1998-2001, Neil McKay
 */

#include "tk.h"

/* Include all the X11 stuff we need. */

#include <X11/Xutil.h>

/* We need this for translating network host names. */

#ifdef HAVE_NET_ERRNO_H
#   include <net/errno.h>
#endif
#include <netinet/in.h>         /* struct in_addr, struct sockaddr_in */
#include <netdb.h>              /* gethostbyname() */

#include "xopInt.h"


/*
 *--------------------------------------------------------------
 *
 * BytesAreZero --
 *
 *	Tests to see if the bytes at the given address are zero.
 *
 * Results:
 *	Returns 1 if the bytes are zero, 1 otherwise.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
BytesAreZero(bytePtr, nBytes)
    char *bytePtr;	/* Bytes to test	*/
    int nBytes;		/* # of bytes to test	*/
{
    int i;

    for(i = 0; i < nBytes; i++) {
	if (bytePtr[i] != 0) {
	    return 0;
	}
    }
    return 1;
}


/*
 *--------------------------------------------------------------
 *
 * HostNameToAddressList --
 *
 *	Converts a host name to a list of numeric IP addresses.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
HostNameToAddressList(interp, name, nAddrPtr, addrListPtr)
    Tcl_Interp *interp;		/* Error messages go here	*/
    char *name;			/* Name to be translated	*/
    int *nAddrPtr;		/* Number of addresses returned	*/
    struct in_addr **addrListPtr;/* Address list		*/
{
    struct hostent *hostEntPtr;
    int i, nAddresses;
    char **charAddrPtr;

    hostEntPtr = gethostbyname(name);
    if (hostEntPtr == NULL) {
	Tcl_AppendResult(interp,
		"No address found for host \"", name, "\"", (char *) NULL);
	return TCL_ERROR;
    }

    /* Find number of addresses returned. */

    nAddresses = 0;
    charAddrPtr =  hostEntPtr->h_addr_list;
    while (!BytesAreZero(*charAddrPtr,hostEntPtr->h_length)) {
	 charAddrPtr++;
	 nAddresses++;
    }
    *nAddrPtr = nAddresses;

    /*
     *@@@ POTENTIAL MEMORY LEAK! do we need to free any of the
     *@@@ hostent struct? If so, make sure you call the right
     *@@@ routine to do it. (Probably "free", NOT ckfree or Tcl_Free.)
     */

    /* Allocate storage and copy addresses. */

    if (nAddresses == 0) {
	*addrListPtr = NULL;
	return TCL_OK;
    }

    *addrListPtr = (struct in_addr *) ckalloc(nAddresses*sizeof(struct in_addr));
    for(charAddrPtr = hostEntPtr->h_addr_list;
		i < nAddresses; i++, charAddrPtr += hostEntPtr->h_length) {
	memcpy(&((*addrListPtr)[i]), charAddrPtr, hostEntPtr->h_length);
    }

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * HostAddressToName --
 *
 *	Converts a numeric host address to a host name.
 *
 * Results:
 *	Returns a string containing the host name. This string
 *	should be freed by "ckfree" after use. If the address
 *	couldn't be converted, NULL is returned.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static char *
HostAddressToName(interp, addressPtr, addressLen)
    Tcl_Interp *interp;	/* Error messages go here.	*/
    char *addressPtr;	/* Pointer to numeric IP addr.	*/
    int addressLen;	/* Length of address		*/
{
    struct hostent *hostEntPtr;
    char *hostName;

    hostEntPtr = gethostbyaddr(addressPtr, addressLen, AF_INET);

    if (hostEntPtr == (struct hostent *) NULL) {
	/*
	 *@@@ Should we just translate the address into a
	 *@@@ dotted quad?
	 */

	hostName = NULL;
    } else {
	hostName = ckalloc(1 + strlen(hostEntPtr->h_name));
	strcpy(hostName, hostEntPtr->h_name);
	/*
	 *@@@ Do we need to free anything returned by gethostbyaddr?
	 *@@@ POTENTIAL MEMORY LEAK! (If we do have to free anything,
	 *@@@ make sure to use the correct storage-freeing routine;
	 *@@@ this is probably "free", NOT ckfree or Tcl_Free.)
	 */
    }

    return hostName;
}


/*
 *--------------------------------------------------------------
 *
 * XopAccessCommand --
 *
 *	Implements the "access" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopAccessCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    if (argc < 4) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" control|hostlist args", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "control") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp,
			"Usage: ", argv[0], " ", argv[1], " ", argv[2],
			" enable|disable", (char *) NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[3], "enable") == 0) {
	    XEnableAccessControl(opts->display);
	    return TCL_OK;
	}

	if (strcmp(argv[3], "disable") == 0) {
	    XDisableAccessControl(opts->display);
	    return TCL_OK;
	}

	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1], " ", argv[2],
		" enable|disable", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "hostlist") == 0) {
	if (argc > 5) {
	    Tcl_AppendResult(interp,
			"Usage: ", argv[0], " ", argv[1], " ", argv[2],
			" add|get|remove ...", (char *) NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[3], "add") == 0) {
	    Tcl_SetResult(interp,
		    "hostlist add command not yet implemented", TCL_STATIC);
	    return TCL_ERROR;
	    /*@@@ Use gethostbyname to convert names to IP addresses @@@*/
	}

	if (strcmp(argv[3], "get") == 0) {
	    Tcl_SetResult(interp,
		    "hostlist get command not yet implemented", TCL_STATIC);
	    return TCL_ERROR;
	    /*@@@ Use gethostbyaddr to convert IP address to name @@@*/
	}

	if (strcmp(argv[3], "remove") == 0) {
	    Tcl_SetResult(interp,
		    "hostlist remove command not yet implemented", TCL_STATIC);
	    return TCL_ERROR;
	    /*@@@ Use gethostbyname to convert names to IP addresses @@@*/
	}

	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1], " ", argv[2],
		" add|get|remove ...", (char *) NULL);
	return TCL_ERROR;
    }

    /* None of the above: error. */

    Tcl_AppendResult(interp,
	    "Usage: ", argv[0], " ", argv[1],
	    " control|hostlist args", (char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopBellCommand --
 *
 *	Implements the "bell" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static char *bellProp[] = {"duration", "frequency", "volume", NULL};
static int bellArgMax[] = {32767, 32767, 100};

int
XopBellCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    XKeyboardState kbState;
    XKeyboardControl kbControls;
    int valueMask;
    int num, bellPercent;
    char junkStr[32];
    int bellPropIdx;

    if (argc < 3 || argc > 4) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" ring|duration|frequency|volume ...", (char *) NULL);
	return TCL_ERROR;
    }

    /* First, handle the "ring" command. */

    if (strcmp(argv[2], "ring") == 0) {
	bellPercent = 100;
	if (argc == 4) {
	    if (Tcl_GetInt(interp, argv[3], &bellPercent) != TCL_OK) {
		return TCL_ERROR;
	    }
	    if (bellPercent < 0) {
		bellPercent = 0;
	    } else if (bellPercent > 100) {
		bellPercent = 100;
	    }
	}
	XBell(opts->display, bellPercent);
	return TCL_OK;
    }

    /*
     * With the "ring" command out of the way, we know that
     * the others set/get bell characteristics. If argc == 3,
     * we're getting characteristics, if argc == 4 we're
     * setting them. First, look up the subcommand name.
     */

    for (bellPropIdx = 0; bellProp[bellPropIdx] != NULL; bellPropIdx++) {
	if (strcmp(argv[2], bellProp[bellPropIdx]) == 0) {
	    break;
	}
    }

    if (bellProp[bellPropIdx] == NULL) {
	Tcl_AppendResult(interp, "unknown bell command \"", argv[2],
		"\"; must be one of:", (char *) NULL);
	for (bellPropIdx = 0; bellProp[bellPropIdx] != NULL; bellPropIdx++) {
	    Tcl_AppendResult(interp, " ",
			bellProp[bellPropIdx], (char *) NULL);
	}

	return TCL_ERROR;
    }

    /* Handle getting characteristics. */

    if (argc == 3) {
	XGetKeyboardControl(opts->display, &kbState);

	switch(bellPropIdx) {
	    case 0:
		num = kbState.bell_duration;
		break;

	    case 1:
		num = kbState.bell_pitch;
		break;

	    case 2:
		num = kbState.bell_percent;
		break;
	}

	sprintf(junkStr, "%1d", num);
	Tcl_SetResult(interp, junkStr, TCL_VOLATILE);

	return TCL_OK;
    }

    /*
     * At this point, we must have argc == 4;
     * handle setting characteristics.
     */

    if (Tcl_GetInt(interp, argv[3], &num) != TCL_OK) {
	return TCL_ERROR;
    }

    if (num < -1) {
	num = -1;
    }
    if (num > bellArgMax[bellPropIdx]) {
	num = bellArgMax[bellPropIdx];
    }

    switch(bellPropIdx) {
	case 0:
	    valueMask = KBBellDuration;
	    kbControls.bell_duration = num;
	    break;

	case 1:
	    valueMask = KBBellPitch;
	    kbControls.bell_pitch = num;
	    break;

	case 2:
	    valueMask = KBBellPercent;
	    kbControls.bell_percent = num;
	    break;
    }

    XChangeKeyboardControl(opts->display, valueMask, &kbControls);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * XopBufferCommand --
 *
 *	Implements the "buffer" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopBufferCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    int bufNo, bufLen;
    char *bufContents;
    char *bufCopy;
    Tk_ErrorHandler handler;
    int gotXError;

    if (argc < 4) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" <bufferNumber> contents|rotate ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    /* Get the buffer number. */

    if (Tcl_GetInt(interp, argv[2], &bufNo) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Do the commands... */

    if (strcmp(argv[3], "rotate") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp,
			"Usage: ", argv[0], " ", argv[1],
			" <bufferNumber> rotate", (char *) NULL);
	    return TCL_ERROR;
	}

	/*
	 * Here's where we actually rotate buffers. Note that
	 * XRotateBuffers will fail if ANY of the cut buffers
	 * has not been set, so we have to check errors. The
	 * XSync call is needed in order to force the rotation
	 * to occur while the proper error handler is installed;
	 * without it, the default error handler will be re-
	 * installed before the actual occurrence of the error
	 * caused by the XRotateBuffers call.
	 */

	gotXError = 0;
	handler = Tk_CreateErrorHandler(opts->display, -1, -1, -1,
		XopXErrorHandler, (ClientData) (&gotXError));
	XRotateBuffers(opts->display, bufNo%8);
	Tk_DeleteErrorHandler(handler);
	XSync(opts->display,False);

	if (gotXError) {
	    Tcl_SetResult(interp, "couldn't rotate buffers", TCL_STATIC);
	    return TCL_ERROR;
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "contents") == 0) {
	if (argc > 5) {
	    Tcl_AppendResult(interp,
			"Usage: ", argv[0], " ", argv[1],
			" <bufferNumber> contents ?string?", (char *) NULL);
	    return TCL_ERROR;
	}

	if (argc == 4) {
	    bufContents = XFetchBuffer(opts->display, &bufLen, bufNo%8);
	    if(bufLen != 0) {
		bufCopy = (char *) ckalloc(1 + bufLen);
		strncpy(bufCopy, bufContents, bufLen);
		bufCopy[bufLen] = '\0';
		Tcl_SetResult(interp, bufCopy, TCL_DYNAMIC);
		XFree(bufContents);
	    }

	    return TCL_OK;
	}

	/* Should have argc == 5. */

	XStoreBuffer(opts->display, argv[4], strlen(argv[4]), bufNo%8);
	return TCL_OK;
    }

    Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
			" <bufferNumber> rotate|contents ...",
		(char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * XopEventCommand --
 *
 *	Implements the "event" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopEventCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    int num;
    char junkStr[32];
    Window recipientWindow;
    XEvent sendEvent;
    long eventMask;
    Tk_ErrorHandler handler;
    int gotXError;
    Status status;

    if (argc < 3) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" numpending|send ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "numpending") == 0) {
	num = XEventsQueued(opts->display, QueuedAlready);
	sprintf(junkStr, "%1d", num);
	Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	return TCL_OK;
    }

    if (strcmp(argv[2], "send") == 0) {
	/* Make sure we have enough arguments. */

	if (argc < 5) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " windowID eventType ?args?",
			(char *) 0);
	    return TCL_ERROR;
	}

	/* Get recipient window ID. */

	if (strlen(argv[3]) == 0) {
	    Tcl_SetResult(interp,
		    "window ID cannot be null", TCL_STATIC);
	    return TCL_ERROR;
	}

	if (strncasecmp(argv[3], "pointerwindow", strlen(argv[3])) == 0) {
	    recipientWindow = PointerWindow;
	} else if (strncasecmp(argv[3], "inputfocus", strlen(argv[3])) == 0) {
	    recipientWindow = InputFocus;
	} else if  (XopGetAndCheckWindowId(interp, opts->display,
				argv[3], &recipientWindow) != TCL_OK) {
	    return TCL_ERROR;
	}

	/* Get the XEvent struct that we'll send. */

	if(XopParseEvent(interp, opts->display, recipientWindow,
		&sendEvent, &eventMask, argc-4, argv+4) != TCL_OK) {
	    return TCL_ERROR;
	}

	/* Send the event. (Handle errors appropriately.) */

	gotXError = 0;
	handler = Tk_CreateErrorHandler(opts->display, -1, -1, -1,
			XopXErrorHandler, (ClientData) (&gotXError));

	status = XSendEvent(opts->display,
			recipientWindow, True, eventMask, &sendEvent);
	Tk_DeleteErrorHandler(handler);
	XSync(opts->display, False);

	if (gotXError || (status == 0)) {
	    Tcl_SetResult(interp, "Error sending event", TCL_STATIC);
	    return TCL_ERROR;
	}

	return TCL_OK;
    }

    Tcl_AppendResult(interp,
	    "Usage: ", argv[0], " ", argv[1],
	    " numpending|send ...",
	    (char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopFontCommand --
 *
 *	Implements the "font" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopFontCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    char *pattern;
    char **fontList;
    int i, fontCount;
    char **fontPath;
    int pathDirCount;

    if (argc < 3) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" listnames|path ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "listnames") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?pattern?", (char *) NULL);
	    return TCL_ERROR;
	}

	pattern = "*";
	if (argc > 3) {
	    pattern = argv[3];
	}

	fontList = XListFonts(opts->display, pattern, 99999, &fontCount);
	for(i = 0; i < fontCount; i++) {
	    Tcl_AppendElement(interp, fontList[i]);
	}
	XFreeFontNames(fontList);
	return TCL_OK;
    }

    if (strcmp(argv[2], "path") == 0) {
	if (argc < 3 || argc > 4) {
	    Tcl_AppendResult(interp,
			"Usage: ", argv[0], " ", argv[1], " ",
			argv[2], " ?fontPath?", (char *) NULL);
	    return TCL_ERROR;
	}

	if (argc == 3) {
	    fontPath = XGetFontPath(opts->display, &pathDirCount);
	    for(i = 0; i < pathDirCount; i++) {
		Tcl_AppendElement(interp, fontPath[i]);
	    }
	    XFreeFontPath(fontPath);
	    return TCL_OK;
	}

	/* Setting font path */

	if (Tcl_SplitList(interp,
			argv[3], &pathDirCount, &fontPath) != TCL_OK) {
	    return TCL_ERROR;
	}

	XSetFontPath(opts->display, fontPath, pathDirCount);
	ckfree((VOID *) fontPath);
	return TCL_OK;
    }

    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		"listnames|path ...", (char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopSavesetCommand --
 *
 *	Implements the "saveset" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopSavesetCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    Window windowID;
    int result;

    if (argc != 4) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" add|remove windowID", (char *) 0);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "add") == 0) {
	result = XopGetAndCheckWindowId(interp,
			opts->display, argv[3], &windowID);
	if(result != TCL_OK) {
	    return TCL_ERROR;
	}

	XAddToSaveSet(opts->display, windowID);
	return TCL_OK;
    }

    if (strcmp(argv[2], "remove") == 0) {
	result = XopGetAndCheckWindowId(interp,
			opts->display, argv[3], &windowID);
	if(result != TCL_OK) {
	    return TCL_ERROR;
	}

	XRemoveFromSaveSet(opts->display, windowID);
	return TCL_OK;
    }

    Tcl_AppendResult(interp, "\"", argv[1],
		"\" subcommand not yet implemented", (char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopScreensaverCommand --
 *
 *	Implements the "screensaver" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopScreensaverCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    int timeout, interval, blanking, exposures;
    char junkStr[32];

    if (argc < 3) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" activate|blanking|exposures|interval|reset|timeout ...",
		(char *) 0);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "activate") == 0) {
	if (argc > 3) {
	    Tcl_AppendResult(interp, "extraneous arguments follow \"",
			argv[2], "\"", (char *) 0);
	    return TCL_ERROR;
	}

	XActivateScreenSaver(opts->display);
	return TCL_OK;
    }

    if (strcmp(argv[2], "blanking") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?blankingPreference?", (char *) 0);
	    return TCL_ERROR;
	}

	/* Get the current settings. */

	XGetScreenSaver(opts->display, &timeout,
				&interval, &blanking, &exposures);

	/* Return the blanking setting, if that was requested. */

	if (argc == 3) {
	    if (blanking == DontPreferBlanking) {
		Tcl_SetResult(interp, "dontprefer", TCL_STATIC);
	    } else if (blanking == PreferBlanking) {
		Tcl_SetResult(interp, "prefer", TCL_STATIC);
	    } else if (blanking == DefaultBlanking) {
		Tcl_SetResult(interp, "default", TCL_STATIC);
	    }

	    return TCL_OK;
	}

	/* Get the desired blanking setting. */

	if (strcmp(argv[3], "dontprefer") == 0) {
		blanking = DontPreferBlanking;
	} else if (strcmp(argv[3], "prefer") == 0) {
		blanking = PreferBlanking;
	} else if (strcmp(argv[3], "default") == 0) {
		blanking = DefaultBlanking;
	} else {
	    Tcl_AppendResult(interp, "unknown blanking preference \"", argv[3],
			"\"; should be prefer, dontprefer, or default",
			(char *) 0);
	    return TCL_ERROR;
	}

	/* Change the screensaver settings. */

	XSetScreenSaver(opts->display, timeout, interval, blanking, exposures);
	return TCL_OK;
    }

    if (strcmp(argv[2], "exposures") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?allowExposures?", (char *) 0);
	    return TCL_ERROR;
	}

	/* Get the current settings. */

	XGetScreenSaver(opts->display, &timeout,
				&interval, &blanking, &exposures);

	/* Return the exposures setting, if that was requested. */

	if (argc == 3) {
	    if (exposures == DontAllowExposures) {
		Tcl_SetResult(interp, "dontallow", TCL_STATIC);
	    } else if (exposures == AllowExposures) {
		Tcl_SetResult(interp, "allow", TCL_STATIC);
	    } else if (exposures == DefaultExposures) {
		Tcl_SetResult(interp, "default", TCL_STATIC);
	    }

	    return TCL_OK;
	}

	/* Get the desired blanking setting. */

	if (strcmp(argv[3], "dontallow") == 0) {
		exposures = DontAllowExposures;
	} else if (strcmp(argv[3], "allow") == 0) {
		exposures = AllowExposures;
	} else if (strcmp(argv[3], "default") == 0) {
		exposures = DefaultExposures;
	} else {
	    Tcl_AppendResult(interp, "unknown exposure setting \"", argv[3],
			"\"; should be allow, dontallow, or default",
			(char *) 0);
	    return TCL_ERROR;
	}

	/* Change the screensaver settings. */

	XSetScreenSaver(opts->display, timeout, interval, blanking, exposures);
	return TCL_OK;
    }

    if (strcmp(argv[2], "interval") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?changeInterval?", (char *) 0);
	    return TCL_ERROR;
	}

	/* Get the current settings. */

	XGetScreenSaver(opts->display, &timeout,
				&interval, &blanking, &exposures);

	/* Return the change_interval setting, if that was requested. */

	if (argc == 3) {
	    sprintf(junkStr, "%1d", interval);
	    Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	    return TCL_OK;
	}

	/* Get the desired screensaver change interval. */

	if (Tcl_GetInt(interp, argv[3], &interval) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (interval < -1) {
	    interval = -1;
	} else if (interval > 32767) {
	    interval = 32767;
	}

	/* Change the screensaver settings. */

	XSetScreenSaver(opts->display, timeout, interval, blanking, exposures);
	return TCL_OK;
    }

    if (strcmp(argv[2], "reset") == 0) {
	if (argc > 3) {
	    Tcl_AppendResult(interp, "extraneous arguments follow \"",
			argv[2], "\"", (char *) 0);
	    return TCL_ERROR;
	}

	XResetScreenSaver(opts->display);
	return TCL_OK;
    }

    if (strcmp(argv[2], "timeout") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?timeoutInterval?", (char *) 0);
	    return TCL_ERROR;
	}

	/* Get the current settings. */

	XGetScreenSaver(opts->display, &timeout,
				&interval, &blanking, &exposures);

	/* Return the timeout setting, if that was requested. */

	if (argc == 3) {
	    sprintf(junkStr, "%1d", timeout);
	    Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	    return TCL_OK;
	}

	/* Get the desired screensaver change interval. */

	if (Tcl_GetInt(interp, argv[3], &timeout) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (timeout < -1) {
	    timeout = -1;
	} else if (timeout > 32767) {
	    timeout = 32767;
	}

	/* Change the screensaver settings. */

	XSetScreenSaver(opts->display, timeout, interval, blanking, exposures);
	return TCL_OK;
    }

    Tcl_AppendResult(interp, "Unknown subcommand \"", argv[2],
		"\" should be activate, blanking, exposures, interval, or reset",
		(char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopServerCommand --
 *
 *	Implements the "server" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopServerCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    char **extList;
    int i, nExtensions;
    char junkStr[32];

    if (argc != 3) {
	Tcl_AppendResult(interp, "Usage: ",
		argv[0], " ", argv[1], " command, where command may be:\n",
		"\textensions grab ungrap vendor vendorrelease",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "extensions") == 0) {
	extList = XListExtensions(opts->display, &nExtensions);
	for(i = 0; i < nExtensions; i++) {
	    Tcl_AppendElement(interp, extList[i]);
	}
	XFreeExtensionList(extList);
	return TCL_OK;
    }

    if (strcmp(argv[2], "grab") == 0) {
	XGrabServer(opts->display);
	return TCL_OK;
    }

    if (strcmp(argv[2], "ungrab") == 0) {
	XUngrabServer(opts->display);
	return TCL_OK;
    }

    if (strcmp(argv[2], "vendor") == 0) {
	Tcl_SetResult(interp, ServerVendor(opts->display), TCL_VOLATILE);
	return TCL_OK;
    }

    if (strcmp(argv[2], "vendorrelease") == 0) {
	sprintf(junkStr, "%1d", VendorRelease(opts->display));
	Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	return TCL_OK;
    }

    Tcl_AppendResult(interp, "Usage: ",
		argv[0], " ", argv[1], " command, where command may be:\n",
		"\textensions grab ungrap vendor vendorrelease",
		(char *) NULL);
    return TCL_ERROR;
}

