/* 
 * wm.c --
 *
 *	This file contains code for xop's window-manager-related
 *	subcommands: wmprops, focus, and colormap.
 */

/*
 * Copyright (C) 1998-2001, Neil McKay
 */

#include "tk.h"

/* Include all the X11 stuff we need. */

#include <X11/Xutil.h>

/* We need this for translating network host names. */

#ifdef HAVE_NET_ERRNO_H
#   include <net/errno.h>
#endif
#include <netinet/in.h>         /* struct in_addr, struct sockaddr_in */
#include <netdb.h>              /* gethostbyname() */

#include "xopInt.h"


/*
 *--------------------------------------------------------------
 *
 * WriteWMHints --
 *
 *	Write the contents of a XWMHints sructure as a Tcl
 *	list.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The interpreter's result string is set.
 *
 *--------------------------------------------------------------
 */

static void
WriteWMHints(interp, hintsPtr)
    Tcl_Interp *interp;	/* Text version of hints goes here	*/
    XWMHints *hintsPtr;	/* WM Hints structure			*/
{
    char junkStr[32];

    if (hintsPtr->flags & InputHint) {
	Tcl_AppendElement(interp, "inputWM");
	if(hintsPtr->input) {
	    Tcl_AppendElement(interp, "1");
	} else {
	    Tcl_AppendElement(interp, "0");
	}
    }

    if (hintsPtr->flags & StateHint) {
	Tcl_AppendElement(interp, "initState");
	if(hintsPtr->initial_state == DontCareState) {
	    Tcl_AppendElement(interp, "DontCare");
	} else if (hintsPtr->initial_state == NormalState) {
	    Tcl_AppendElement(interp, "Normal");
	} else if (hintsPtr->initial_state == ZoomState) {
	    Tcl_AppendElement(interp, "Zoom");
	} else if (hintsPtr->initial_state == IconicState) {
	    Tcl_AppendElement(interp, "Iconic");
	} else if (hintsPtr->initial_state == InactiveState) {
	    Tcl_AppendElement(interp, "Inactive");
	}
    }

    if (hintsPtr->flags & IconWindowHint) {
	Tcl_AppendElement(interp, "iconWindowId");
	sprintf(junkStr, "0x%x", (unsigned int) (hintsPtr->icon_window));
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & IconPixmapHint) {
	Tcl_AppendElement(interp, "iconPixmapId");
	sprintf(junkStr, "0x%x", (unsigned int) (hintsPtr->icon_pixmap));
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & IconPositionHint) {
	Tcl_AppendElement(interp, "iconPosition");
	sprintf(junkStr, "%1d %1d", hintsPtr->icon_x, hintsPtr->icon_y);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & IconMaskHint) {
	Tcl_AppendElement(interp, "iconMaskId");
	sprintf(junkStr, "0x%x", (unsigned int) (hintsPtr->icon_mask));
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & WindowGroupHint) {
	Tcl_AppendElement(interp, "groupWindowId");
	sprintf(junkStr, "0x%x", (unsigned int) (hintsPtr->window_group));
	Tcl_AppendElement(interp, junkStr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * WriteWMNormalHints --
 *
 *	Write the contents of a XSizeHints sructure as a Tcl
 *	list.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The interpreter's result string is set.
 *
 *--------------------------------------------------------------
 */

static void
WriteWMNormalHints(interp, hintsPtr, mask)
    Tcl_Interp *interp;	/* Text version of hints goes here	*/
    XSizeHints *hintsPtr;	/* WM SizeHints structure	*/
    unsigned long mask;		/* Mask showing what's valid	*/
{
    char junkStr[32];

    if (hintsPtr->flags & USPosition) {
	Tcl_AppendElement(interp, "userPosition");
	sprintf(junkStr, "%1d %1d", hintsPtr->x, hintsPtr->y);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & USSize) {
	Tcl_AppendElement(interp, "userSize");
	sprintf(junkStr, "%1d %1d", hintsPtr->width, hintsPtr->height);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PPosition) {
	Tcl_AppendElement(interp, "progPosition");
	sprintf(junkStr, "%1d %1d", hintsPtr->x, hintsPtr->y);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PSize) {
	Tcl_AppendElement(interp, "progSize");
	sprintf(junkStr, "%1d %1d", hintsPtr->width, hintsPtr->height);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PMinSize) {
	Tcl_AppendElement(interp, "minSize");
	sprintf(junkStr, "%1d %1d", hintsPtr->min_width, hintsPtr->min_height);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PMaxSize) {
	Tcl_AppendElement(interp, "maxSize");
	sprintf(junkStr, "%1d %1d", hintsPtr->max_width, hintsPtr->max_height);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PResizeInc) {
	Tcl_AppendElement(interp, "resizeInc");
	sprintf(junkStr, "%1d %1d", hintsPtr->width_inc, hintsPtr->height_inc);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PAspect) {
	Tcl_AppendElement(interp, "minAspect");
	sprintf(junkStr, "%1d %1d",
		hintsPtr->min_aspect.x, hintsPtr->min_aspect.y);
	Tcl_AppendElement(interp, junkStr);
	Tcl_AppendElement(interp, "maxAspect");
	sprintf(junkStr, "%1d %1d",
		hintsPtr->max_aspect.x, hintsPtr->max_aspect.y);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PBaseSize) {
	Tcl_AppendElement(interp, "baseSize");
	sprintf(junkStr, "%1d %1d",
		hintsPtr->base_width, hintsPtr->base_height);
	Tcl_AppendElement(interp, junkStr);
    }

    if (hintsPtr->flags & PWinGravity) {
	Tcl_AppendElement(interp, "winGravity");
	if(hintsPtr->win_gravity == CenterGravity) {
	    Tcl_AppendElement(interp, "c");
	} else if (hintsPtr->win_gravity == EastGravity) {
	    Tcl_AppendElement(interp, "e");
	} else if (hintsPtr->win_gravity == NorthEastGravity) {
	    Tcl_AppendElement(interp, "ne");
	} else if (hintsPtr->win_gravity == NorthGravity) {
	    Tcl_AppendElement(interp, "n");
	} else if (hintsPtr->win_gravity == NorthWestGravity) {
	    Tcl_AppendElement(interp, "nw");
	} else if (hintsPtr->win_gravity == SouthEastGravity) {
	    Tcl_AppendElement(interp, "se");
	} else if (hintsPtr->win_gravity == SouthGravity) {
	    Tcl_AppendElement(interp, "s");
	} else if (hintsPtr->win_gravity == SouthWestGravity) {
	    Tcl_AppendElement(interp, "sw");
	} else if (hintsPtr->win_gravity == WestGravity) {
	    Tcl_AppendElement(interp, "w");
	}
    }
}


/*
 *--------------------------------------------------------------
 *
 * XopColormapCommand --
 *
 *	Implements the "colormap" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopColormapCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    if (argc != 3) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" install|uninstall ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "install") == 0) {
	XInstallColormap(opts->display, opts->colormap);
	return TCL_OK;
    }

    if (strcmp(argv[2], "uninstall") == 0) {
	XUninstallColormap(opts->display, opts->colormap);
	return TCL_OK;
    }

    Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" get|install|uninstall ...",
		(char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * XopFocusCommand --
 *
 *	Implements the "focus" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopFocusCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    Window windowID;
    int revertTo;
    int numArgs;
    char **infoList;
    int result;
    Tk_ErrorHandler handler;
    int gotXError;

    if (argc < 2 || argc > 3) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" focusInfo", (char *) 0);
	return TCL_OK;
    }

    if (argc == 2) {
	XGetInputFocus(opts->display, &windowID, &revertTo);
	if (windowID == None) {
	    Tcl_AppendElement(interp, "None");
	} else if (windowID == PointerRoot) {
	    Tcl_AppendElement(interp, "PointerRoot");
	} else {
	    XopAppendDrawableId(interp, (Drawable) windowID);
	}

	if (revertTo == RevertToParent) {
	     Tcl_AppendElement(interp, "Parent");
	} else if (revertTo == RevertToPointerRoot) {
	     Tcl_AppendElement(interp, "PointerRoot");
	} else if (revertTo == RevertToNone) {
	     Tcl_AppendElement(interp, "None");
	}

	return TCL_OK;
    }

    /* Setting the focus window; parse the focusInfo arg. */

    if (Tcl_SplitList(interp, argv[2], &numArgs, &infoList) != TCL_OK) {
	return TCL_ERROR;
    }

    if (numArgs != 2) {
	ckfree((VOID *) infoList);
	Tcl_AppendResult(interp,
		"focusInfo must be a two-element list: ",
		"a window ID and a \"revert to\" indicator",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(infoList[0],"None") == 0) {
	windowID = None;
    } else if (strcmp(infoList[0],"PointerRoot") == 0) {
	windowID = PointerRoot;
    } else {
	result = XopGetAndCheckWindowId(interp,
			opts->display, infoList[0], &windowID);
	if(result != TCL_OK) {
	    ckfree((VOID *) infoList);
	    Tcl_AppendResult(interp,
			"first element of focusInfo must be ",
			"\"None\", \"PointerRoot\", or a valid X window ID",
			(char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (strcmp(infoList[1],"None") == 0) {
	revertTo = RevertToNone;
    } else if (strcmp(infoList[1],"PointerRoot") == 0) {
	revertTo = RevertToPointerRoot;
    } else if (strcmp(infoList[1],"Parent") == 0) {
	revertTo = RevertToParent;
    } else {
	ckfree((VOID *) infoList);
	Tcl_AppendResult(interp,
		"second element of focusInfo must be ",
		"\"None\", \"PointerRoot\", or \"Parent\"",
		(char *) NULL);
	return TCL_ERROR;
    }
    ckfree((VOID *) infoList);

    /* Set the focus window, and handle errors. */

    gotXError = 0;
    handler = Tk_CreateErrorHandler(opts->display, -1, -1, -1,
		XopXErrorHandler, (ClientData) (&gotXError));
    XSetInputFocus(opts->display, windowID, revertTo, CurrentTime);
    Tk_DeleteErrorHandler(handler);
    XSync(opts->display, False);

    if (gotXError) {
	Tcl_SetResult(interp, "Couldn't set focus", TCL_STATIC);
	return TCL_ERROR;
    }

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * XopWMPropsCommand --
 *
 *	Implements the "wmprops" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopWMPropsCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    Window windowId;
    Status status;
    XTextProperty textProp;
    Window *winList;
    int i, nWindows;
    char **argvRet;
    int argcRet;
    Atom Atom_WM_TAKE_FOCUS, Atom_WM_SAVE_YOURSELF, Atom_WM_DELETE_WINDOW;
    int atomCount;
    Atom *atomList;
    XSizeHints normalHints;
    long normalHintsMask;
    XWMHints *hintsPtr;

    if (argc != 4) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" <windowID> prop", (char *) NULL);
	return TCL_ERROR;
    }

    /* Get the window ID. */

    if(XopGetAndCheckWindowId(interp, opts->display,
			argv[2], (Drawable *) (&windowId)) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Get properties... */

    if (strcmp(argv[3], "clientmachine") == 0) {
	status = XGetWMClientMachine(opts->display, windowId, &textProp);
	if (status == 0) {
	    Tcl_SetResult(interp, "", TCL_STATIC);
	} else {
	    Tcl_SetResult(interp, textProp.value, TCL_VOLATILE);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "colormapwindows") == 0) {
	status = XGetWMColormapWindows(opts->display, windowId,
			&winList, &nWindows);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    for(i = 0; i < nWindows; i++) {
		XopAppendDrawableId(interp, (Drawable) (winList[i]));
	    }
	    XFree(winList);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "command") == 0) {
	status = XGetCommand(opts->display, windowId,
			&argvRet, &argcRet);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    for(i = 0; i < argcRet; i++) {
		Tcl_AppendElement(interp, argvRet[i]);
	    }
	    XFreeStringList(argvRet);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "hints") == 0) {
	hintsPtr = XGetWMHints(opts->display, windowId);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (hintsPtr != (XWMHints *) NULL) {
	    WriteWMHints(interp, hintsPtr);
	    XFree(hintsPtr);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "iconname") == 0) {
	status = XGetWMIconName(opts->display, windowId, &textProp);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    Tcl_SetResult(interp, textProp.value, TCL_VOLATILE);
	    XFree(textProp.value);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "name") == 0) {
	status = XGetWMName(opts->display, windowId, &textProp);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    Tcl_SetResult(interp, textProp.value, TCL_VOLATILE);
	    XFree(textProp.value);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "normalhints") == 0) {
	status = XGetWMNormalHints(opts->display, windowId,
					&normalHints, &normalHintsMask);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    WriteWMNormalHints(interp, &normalHints, normalHintsMask);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "protocols") == 0) {
	status = XGetWMProtocols(opts->display, windowId,
					&atomList, &atomCount);
	Tcl_SetResult(interp, "", TCL_STATIC);
	if (status != 0) {
	    Atom_WM_TAKE_FOCUS = XInternAtom(opts->display,
						"WM_TAKE_FOCUS", 0);
	    Atom_WM_SAVE_YOURSELF = XInternAtom(opts->display,
						"WM_SAVE_YOURSELF", 0);
	    Atom_WM_DELETE_WINDOW = XInternAtom(opts->display,
						"WM_DELETE_WINDOW", 0);
	    for(i = 0; i < atomCount; i++) {
		if (atomList[i] == Atom_WM_TAKE_FOCUS) {
		    Tcl_AppendElement(interp, "WM_TAKE_FOCUS");
		} else if (atomList[i] == Atom_WM_SAVE_YOURSELF) {
		    Tcl_AppendElement(interp, "WM_SAVE_YOURSELF");
		} else if (atomList[i] == Atom_WM_DELETE_WINDOW) {
		    Tcl_AppendElement(interp, "WM_DELETE_WINDOW");
		}
	    }
	    XFree(atomList);
	}
	return TCL_OK;
    }

    if (strcmp(argv[3], "sizehints") == 0) {
	/*@@@ N.Y.I. @@@*/

	Tcl_AppendResult(interp, argv[3],
		" not yet implemented", (char *) NULL);
	return TCL_ERROR;
    }

    /* Don't know what this is. */

    Tcl_AppendResult(interp, "Unknown wmprops command \"", argv[3],
		"\"; valid commands are clientmachine,\n",
		"colormapwindows, hints, iconname, name, normalhints\n",
		"protocols, sizehints", (char *) NULL);
    return TCL_ERROR;
}

