/* 
 * tkXop.c --
 *
 *	This module implements the "Xop" command for the Tk
 *	toolkit.  The Xop command gives Tk access to various
 *	low-level utilities in the X window system; these are
 *	mainly useful for writing window manager scripts.
 *
 */

/*
 * Copyright (C) 1998-2001, Neil McKay
 */

#include "tk.h"

/* Include all the X11 stuff we need. */

#include <X11/Xutil.h>

/* We need this for translating network host names. */

#ifdef HAVE_NET_ERRNO_H
#   include <net/errno.h>
#endif
#include <netinet/in.h>         /* struct in_addr, struct sockaddr_in */
#include <netdb.h>              /* gethostbyname() */

#include "xopInt.h"


/*
 *--------------------------------------------------------------
 *
 * ProcessOpts --
 *
 *	Process options for the Xop command. This
 *	is a convenience procedure for processing generic
 *	options for the Xop command. The arguments take to
 *	the Xop command take the form
 *
 *		?option-value pairs? commandClass args
 *
 *	The procedure will return TCL_OK if the options (if any)
 *	parse correctly.
 *
 * Results:
 *	Returns a standard Tcl result
 *
 * Side effects:
 *	*xopOptionsPtr gets set. The argument list will be
 *	shifted to eliminate the options, if present, and 
 *	*argcPtr will be decremented accordingly.
 *
 *--------------------------------------------------------------
 */

static int
ProcessOpts(interp, argcPtr, argv, xopOptionsPtr)
    Tcl_Interp *interp;		/* Error messages go here	*/
    int *argcPtr;		/* Number of arguments		*/
    char **argv;		/* Argument list		*/
    XopOptions *xopOptionsPtr;	/* Pointer to structure where
				 * option values are stored.	*/
{
    int result;
    int i,j,c;
    int screenNumber;
    Tk_Window mainWin;
    Tk_Window displayWin;

    /* Get the default display. */

    mainWin = Tk_MainWindow(interp);
    if (mainWin == NULL) {
	Tcl_SetResult(interp,
		"No main window for this interpreter", TCL_STATIC);
	return(TCL_ERROR);
    }

    displayWin = Tk_NameToWindow(interp, ".", mainWin);
    if (displayWin == NULL) {
	return(TCL_ERROR);
    }
    xopOptionsPtr->display = Tk_Display(displayWin);

    /* Set other default option values. */

    xopOptionsPtr->colormap = None;
    xopOptionsPtr->screenNumber = 0;

    /* Read the options, if any, and shift the argument list. */

    while(*argcPtr >= 2 && argv[1][0] == '-') {
	if (*argcPtr <= 2) {
	    Tcl_AppendResult(interp,
		    "Missing argument after \"", argv[1],"\"", TCL_STATIC);
	    return(TCL_ERROR);
	}

	c = argv[1][1];
	if (c == 'c' && !strcmp(argv[1],"-colormap")) {
	    result = XopGetAndCheckColormapId(interp,
				xopOptionsPtr->display, argv[2],
				&(xopOptionsPtr->colormap));
	    if (result != TCL_OK) {
		return TCL_ERROR;
	    }
	} else if (c == 'd' && !strcmp(argv[1],"-displayof")) {
	    displayWin = Tk_NameToWindow(interp, argv[2], mainWin);
	    if (displayWin == NULL) {
		return(TCL_ERROR);
	    }
	    xopOptionsPtr->display = Tk_Display(displayWin);
	} else if (c == 's' && !strcmp(argv[1],"-screennumber")) {
	    if (Tcl_GetInt(interp, argv[2], &screenNumber) != TCL_OK) {
		return(TCL_ERROR);
	    }
	} else {
	    Tcl_AppendResult(interp, "Unknown option \"",
			argv[1], "\"", (char *) 0);
	}

	/* Shift the argument list down, to get rid of the option. */

	for(i = 1, j = 3; j < *argcPtr; i++, j++) {
	    argv[i] = argv[j];
	}
	(*argcPtr) -= 2;
    }

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * XopCmd --
 *
 *	Implements the "xop" command. See the xop man page
 *	for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
XopCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Information about rootwin widget. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int result = TCL_OK;
    size_t length;
    int c;
    XopOptions opts;

    if (argc < 2) {
	Tcl_AppendResult(interp, "Usage: ",
		argv[0], " ?options? command_class ?arg arg ...?\n",
		"where command_class may be access, buffer, colormap,\n",
		"drawable, event, focus, font, keyboard, pointer,\n",
		"saveset, screensaver, server, window, or wmprops",
		(char *) NULL);
	return TCL_ERROR;
    }

    /* Preprocess the options. */

    result = ProcessOpts(interp, &argc, argv, &opts);
    if(result != TCL_OK) {
	return TCL_ERROR;
    }

    c = argv[1][0];
    length = strlen(argv[1]);
    if (c == 'a') {
	if (strcmp(argv[1],"access") == 0) {
	    return(XopAccessCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'b') {
	if (strcmp(argv[1],"bell") == 0) {
	    return(XopBellCommand(interp, argc, argv, &opts));
	}
	if (strcmp(argv[1],"buffer") == 0) {
	    return(XopBufferCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'c') {
	if (strcmp(argv[1],"colormap") == 0) {
	    return(XopColormapCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'd') {
	if (strcmp(argv[1],"drawable") == 0) {
	    return(XopDrawableCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'e') {
	if (strcmp(argv[1],"event") == 0) {
	    return(XopEventCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'f') {
	if (strcmp(argv[1],"focus") == 0) {
	    return(XopFocusCommand(interp, argc, argv, &opts));
	}
	if (strcmp(argv[1],"font") == 0) {
	    return(XopFontCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'k') {
	if (strcmp(argv[1],"keyboard") == 0) {
	    return(XopKeyboardCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'p') {
	if (strcmp(argv[1],"pointer") == 0) {
	    return(XopPointerCommand(interp, argc, argv, &opts));
	}
    } else if (c == 's') {
	if (strcmp(argv[1],"saveset") == 0) {
	    return(XopSavesetCommand(interp, argc, argv, &opts));
	}
	if (strcmp(argv[1],"screensaver") == 0) {
	    return(XopScreensaverCommand(interp, argc, argv, &opts));
	}
	if (strcmp(argv[1],"server") == 0) {
	    return(XopServerCommand(interp, argc, argv, &opts));
	}
    } else if (c == 'w') {
	if (strcmp(argv[1],"window") == 0) {
	    return(XopWindowCommand(interp, argc, argv, &opts));
	}
	if (strcmp(argv[1],"wmprops") == 0) {
	    return(XopWMPropsCommand(interp, argc, argv, &opts));
	}
    }

    Tcl_AppendResult(interp,
		"Usage: ",
		argv[0], " ?options? command_class ?arg arg ...?\n",
		"where command_class may be access, bell, buffer, colormap,\n",
		"drawable, event, focus, font, keyboard, pointer,\n",
		"saveset, screensaver, server, window, or wmprops",
		(char *) NULL);
    return TCL_ERROR;
}

int
Xop_Init(interp)
    Tcl_Interp *interp;
{
    if(Tcl_PkgRequire(interp, "Tk", "8.0", 0) == NULL) {
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, "xop", XopCmd, NULL, NULL);
    Tcl_PkgProvide(interp, "Xop", VERSION);

    return TCL_OK;
}

